@@ -5,7 +5,7 @@ import glob
 import json
 import os
 import sys
-from collections import defaultdict
+from collections import defaultdict, OrderedDict
 from datetime import datetime
 from functools import partial as curry
 
@@ -32,7 +32,7 @@ from . import (
     line,
     lint,
     markdown,
-    mdnspeclinks,
+    mdn,
     messages as m,
     metadata,
     refs,
@@ -53,9 +53,9 @@ class Spec:
         self,
         inputFilename: str,
         debug: bool = False,
-        token: t.Optional[str] = None,
+        token: str | None = None,
         lineNumbers: bool = False,
-        fileRequester: t.Optional[retrieve.DataFileRequester] = None,
+        fileRequester: retrieve.DataFileRequester | None = None,
         testing: bool = False,
     ):
         catchArgparseBug(inputFilename)
@@ -72,9 +72,9 @@ class Spec:
             )
             return
         self.inputSource: InputSource.InputSource = InputSource.inputFromName(inputFilename, chroot=constants.chroot)
-        self.transitiveDependencies: t.Set[InputSource.InputSource] = set()
+        self.transitiveDependencies: set[InputSource.InputSource] = set()
         self.debug: bool = debug
-        self.token: t.Optional[str] = token
+        self.token: str | None = token
         self.testing: bool = testing
         self.dataFile: retrieve.DataFileRequester
         if fileRequester is None:
@@ -82,33 +82,35 @@ class Spec:
         else:
             self.dataFile = fileRequester
 
-        self.lines: t.List[line.Line] = []
+        self.lines: list[line.Line] = []
         self.valid = self.initializeState()
 
     def initializeState(self) -> bool:
-        self.normativeRefs: t.Dict[str, biblio.BiblioEntry] = {}
-        self.informativeRefs: t.Dict[str, biblio.BiblioEntry] = {}
+        self.normativeRefs: dict[str, biblio.BiblioEntry] = {}
+        self.informativeRefs: dict[str, biblio.BiblioEntry] = {}
         self.refs: refs.ReferenceManager = refs.ReferenceManager(fileRequester=self.dataFile, testing=self.testing)
-        self.externalRefsUsed: t.Any = defaultdict(lambda: defaultdict(dict))
+        self.externalRefsUsed: refs.ExternalRefsManager = refs.ExternalRefsManager()
 
         self.md: metadata.MetadataManager
-        self.mdBaseline: t.Optional[metadata.MetadataManager] = metadata.MetadataManager()
-        self.mdDocument: t.Optional[metadata.MetadataManager] = None
-        self.mdCommandLine: t.Optional[metadata.MetadataManager] = metadata.MetadataManager()
-        self.mdDefaults: t.Optional[metadata.MetadataManager] = None
-        self.mdOverridingDefaults: t.Optional[metadata.MetadataManager] = None
+        self.mdBaseline: metadata.MetadataManager | None = metadata.MetadataManager()
+        self.mdDocument: metadata.MetadataManager | None = None
+        self.mdCommandLine: metadata.MetadataManager | None = metadata.MetadataManager()
+        self.mdDefaults: metadata.MetadataManager | None = None
+        self.mdOverridingDefaults: metadata.MetadataManager | None = None
 
-        self.typeExpansions: t.Dict[str, str] = {}
+        self.typeExpansions: dict[str, str] = {}
+
+        self.cachedLinksFromHref: OrderedDict[str, list[t.ElementT]] = OrderedDict()
+        self.cachedClassTests: dict[tuple[str, str], bool] = {}
+        self.cachedNormativeEls: dict[t.ElementT, bool] = {}
 
         defaultMacro: t.Callable[[], str] = lambda: "???"
         self.macros: t.DefaultDict[str, str] = defaultdict(defaultMacro)
 
-        self.canIUse: t.Optional[caniuse.CanIUseManager] = None
-        self.mdnSpecLinks: t.Any = {}
         self.widl: widlparser.Parser = idl.getParser()
 
-        self.testSuites: t.Dict[str, testsuite.TestSuite] = fetchTestSuites(self.dataFile)
-        self.languages: t.Dict[str, language.Language] = fetchLanguages(self.dataFile)
+        self.testSuites: dict[str, testsuite.TestSuite] = fetchTestSuites(self.dataFile)
+        self.languages: dict[str, language.Language] = fetchLanguages(self.dataFile)
 
         self.extraStyles: t.DefaultDict[str, str] = defaultdict(str)
         self.extraStyles["style-colors"] = styleColors
@@ -134,7 +136,7 @@ class Spec:
 
         return True
 
-    def recordDependencies(self, *inputSources):
+    def recordDependencies(self, *inputSources: InputSource.InputSource) -> None:
         self.transitiveDependencies.update(inputSources)
 
     def preprocess(self) -> Spec:
@@ -183,7 +185,7 @@ class Spec:
         extensions.load(self)
 
         # Initialize things
-        self.refs.initializeRefs(self)
+        self.refs.initializeRefs(doc=self, datablocks=datablocks)
         self.refs.initializeBiblio()
 
         # Deal with further <pre> blocks, and markdown
@@ -204,8 +206,12 @@ class Spec:
 
         # Build the document
         self.document = h.parseDocument(self.html)
-        self.head = t.cast("t.ElementT", h.find("head", self))
-        self.body = t.cast("t.ElementT", h.find("body", self))
+        headEl = h.find("head", self)
+        bodyEl = h.find("body", self)
+        assert headEl is not None
+        assert bodyEl is not None
+        self.head = headEl
+        self.body = bodyEl
         u.correctFrontMatter(self)
         includes.processInclusions(self)
         metadata.parseDoc(self)
@@ -271,7 +277,7 @@ class Spec:
         boilerplate.addAnnotations(self)
         boilerplate.removeUnwantedBoilerplate(self)
         # Add MDN panels after all IDs/anchors have been added
-        mdnspeclinks.addMdnPanels(self)
+        mdn.addMdnPanels(self)
         highlight.addSyntaxHighlighting(self)
         boilerplate.addBikeshedBoilerplate(self)
         fingerprinting.addTrackingVector(self)
@@ -309,7 +315,7 @@ class Spec:
 
         return self
 
-    def serialize(self) -> t.Optional[str]:
+    def serialize(self) -> str | None:
         try:
             rendered = h.Serializer(self.md.opaqueElements, self.md.blockElements).serialize(self.document)
         except Exception as e:
@@ -318,7 +324,7 @@ class Spec:
         rendered = u.finalHackyCleanup(rendered)
         return rendered
 
-    def fixMissingOutputFilename(self, outputFilename: t.Optional[str]) -> str:
+    def fixMissingOutputFilename(self, outputFilename: str | None) -> str:
         if outputFilename is None:
             # More sensible defaults!
             if isinstance(self.inputSource, InputSource.TarInputSource):
@@ -333,7 +339,7 @@ class Spec:
                 outputFilename = "-"
         return outputFilename
 
-    def finish(self, outputFilename: t.Optional[str] = None, newline: t.Optional[str] = None) -> None:
+    def finish(self, outputFilename: str | None = None, newline: str | None = None) -> None:
         catchArgparseBug(outputFilename)
         self.printResultMessage()
         outputFilename = self.fixMissingOutputFilename(outputFilename)
@@ -365,7 +371,7 @@ class Spec:
             m.success("Successfully generated, with warnings")
             return
 
-    def watch(self, outputFilename: t.Optional[str], port: int = None, localhost: bool = False) -> None:
+    def watch(self, outputFilename: str | None, port: int | None = None, localhost: bool = False) -> None:
         import time
 
         outputFilename = self.fixMissingOutputFilename(outputFilename)
@@ -382,7 +388,7 @@ class Spec:
             import threading
 
             class SilentServer(http.server.SimpleHTTPRequestHandler):
-                def log_message(self, format, *args):
+                def log_message(self, format: t.Any, *args: t.Any) -> None:
                     pass
 
             socketserver.TCPServer.allow_reuse_address = True
@@ -426,7 +432,7 @@ class Spec:
         except Exception as e:
             m.die(f"Something went wrong while watching the file:\n{e}")
 
-    def fixText(self, text: str, moreMacros: t.Dict[str, str] = None) -> str:
+    def fixText(self, text: str, moreMacros: dict[str, str] | None = None) -> str:
         # Do several textual replacements that need to happen *before* the document is parsed as h.
 
         # If markdown shorthands are on, remove all `foo`s while processing,
@@ -446,7 +452,7 @@ class Spec:
         if "css" in self.md.markupShorthands:
             textFunctor = textFunctor.map(h.replaceAwkwardCSSShorthands)
 
-        return textFunctor.extract()
+        return t.cast(str, textFunctor.extract())
 
     def printTargets(self) -> None:
         m.p("Exported terms:")
@@ -481,7 +487,7 @@ def printDone() -> None:
         m.p("")
 
 
-def findImplicitInputFile() -> t.Optional[str]:
+def findImplicitInputFile() -> str | None:
     """
     Find what input file the user *probably* wants to use,
     by scanning the current folder.
@@ -508,7 +514,7 @@ def findImplicitInputFile() -> t.Optional[str]:
     return None
 
 
-def catchArgparseBug(string: t.Optional[str]) -> bool:
+def catchArgparseBug(string: str | None) -> bool:
     # Argparse has had a long-standing bug
     # https://bugs.python.org/issue22433
     # about spaces in the values of unknown optional arguments
@@ -526,11 +532,11 @@ def catchArgparseBug(string: t.Optional[str]) -> bool:
     return True
 
 
-def fetchTestSuites(dataFile: retrieve.DataFileRequester) -> t.Dict[str, testsuite.TestSuite]:
+def fetchTestSuites(dataFile: retrieve.DataFileRequester) -> dict[str, testsuite.TestSuite]:
     return {k: testsuite.TestSuite(**v) for k, v in json.loads(dataFile.fetch("test-suites.json", str=True)).items()}
 
 
-def fetchLanguages(dataFile: retrieve.DataFileRequester) -> t.Dict[str, language.Language]:
+def fetchLanguages(dataFile: retrieve.DataFileRequester) -> dict[str, language.Language]:
     return {
         k: language.Language(v["name"], v["native-name"])
         for k, v in json.loads(dataFile.fetch("languages.json", str=True)).items()
