@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import dataclasses
 import itertools
 import re
@@ -9,7 +11,7 @@ else:
     from . import config, messages as m, t
 
 if t.TYPE_CHECKING:
-    Characters: t.TypeAlias = t.Dict[str, t.List[str]]
+    Characters: t.TypeAlias = dict[str, list[str]]
     T = t.TypeVar("T")
     U = t.TypeVar("U")
 
@@ -81,8 +83,8 @@ class Font:
         self.metadata, lines = parseMetadata(lines)
         self.characters = parseCharacters(self.metadata, lines)
 
-    def write(self, text):
-        output = [""] * self.metadata["height"]
+    def write(self, text: str) -> list[str]:
+        output = [""] * self.metadata.height
         for letterIndex, letter in enumerate(text):
             if letter in self.characters:
                 for i, line in enumerate(self.characters[letter]):
@@ -95,7 +97,7 @@ class Font:
         return output
 
 
-def parseMetadata(lines: t.List[str]) -> t.Tuple[FontMetadata, t.List[str]]:
+def parseMetadata(lines: list[str]) -> tuple[FontMetadata, list[str]]:
     # Each metadata line is of the form "key: value".
     # First line that's not of that form ends the metadata.
     # Returns the parsed metadata, and the non-metadata lines
@@ -124,14 +126,14 @@ def parseMetadata(lines: t.List[str]) -> t.Tuple[FontMetadata, t.List[str]]:
     return md, lines[i:]
 
 
-def parseCharacters(md: FontMetadata, lines: t.List[str]) -> Characters:
+def parseCharacters(md: FontMetadata, lines: list[str]) -> Characters:
     import string
 
     height = md.height
     characters: Characters = {}
     characters[" "] = [" " * md.spaceWidth] * height
     for bigcharlines in grouper(lines, height + 1, ""):
-        littlechar = t.cast(str, bigcharlines[0][0])
+        littlechar = bigcharlines[0][0]
         bigchar = [line.strip("\n") for line in bigcharlines[1:]]
         width = max(len(line) for line in bigchar)
         for i, line in enumerate(bigchar):
@@ -148,30 +150,30 @@ def parseCharacters(md: FontMetadata, lines: t.List[str]) -> Characters:
     return characters
 
 
-def replaceComments(font: Font, inputFilename: t.Optional[str] = None, outputFilename: t.Optional[str] = None) -> None:
+def replaceComments(font: Font, inputFilename: str | None = None, outputFilename: str | None = None) -> None:
     lines, inputFilename = getInputLines(inputFilename)
-    replacements = []
+    replacements: list[tuple[int, list[str]]] = []
     for i, line in enumerate(lines):
         match = re.match(r"\s*<!--\s*Big Text:\s*(\S.*)-->", line)
         if match:
             newtext = ["<!--\n"] + font.write(match.group(1).strip()) + ["-->\n"]
-            replacements.append({"line": i, "content": newtext})
+            replacements.append((i, newtext))
     for r in reversed(replacements):
-        lines[r["line"] : r["line"] + 1] = r["content"]
+        lines[r[0] : r[0] + 1] = r[1]
     writeOutputLines(outputFilename, inputFilename, lines)
 
 
 # Some utility functions
 
 
-def grouper(iterable: t.Sequence[T], n: int, fillvalue: U) -> t.Generator[t.List[t.Union[T, U]], None, None]:
+def grouper(iterable: t.Sequence[T], n: int, fillvalue: U) -> t.Generator[list[T | U], None, None]:
     "Collect data into fixed-length chunks or blocks"
     # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx
     args = [iter(iterable)] * n
-    return (list(t.cast("t.Iterable[t.Union[T,U]]", x)) for x in itertools.zip_longest(fillvalue=fillvalue, *args))
+    return (list(t.cast("t.Iterable[T|U]", x)) for x in itertools.zip_longest(fillvalue=fillvalue, *args))
 
 
-def getInputLines(inputFilename: t.Optional[str]) -> t.Tuple[t.List[str], str]:
+def getInputLines(inputFilename: str | None) -> tuple[list[str], str]:
     if inputFilename is None:
         # Default to looking for a *.bs file.
         # Otherwise, look for a *.src.html file.
@@ -199,7 +201,7 @@ def getInputLines(inputFilename: t.Optional[str]) -> t.Tuple[t.List[str], str]:
     return lines, inputFilename
 
 
-def writeOutputLines(outputFilename: t.Optional[str], inputFilename: str, lines: t.List[str]) -> None:
+def writeOutputLines(outputFilename: str | None, inputFilename: str, lines: list[str]) -> None:
     if outputFilename is None:
         outputFilename = inputFilename
     try:
