@@ -1,12 +1,18 @@
 from __future__ import annotations
 
+import dataclasses
 import json
 import random
 import re
 from collections import defaultdict
 
-from .. import biblio, config, constants, datablocks, h, messages as m, retrieve, t
-from . import utils, source
+from .. import biblio, config, constants, h, messages as m, retrieve, t
+from . import utils, source, headingdata, wrapper
+
+if t.TYPE_CHECKING:
+    from .wrapper import RefWrapper
+
+    SpecDataT: t.TypeAlias = dict[str, str]
 
 
 class ReferenceManager:
@@ -35,85 +41,91 @@ class ReferenceManager:
 
     def __init__(
         self,
-        defaultStatus: t.Optional[str] = None,
-        fileRequester: t.Optional[retrieve.DataFileRequester] = None,
+        defaultStatus: str | None = None,
+        fileRequester: t.DataFileRequester | None = None,
         testing: bool = False,
     ):
+        self.dataFile: t.DataFileRequester
         if fileRequester is None:
             self.dataFile = retrieve.defaultRequester
         else:
             self.dataFile = fileRequester
 
-        self.testing = testing
+        self.testing: bool = testing
 
         # Dict of {spec vshortname => spec data}
-        self.specs: t.Dict[str, t.Dict] = dict()
+        self.specs: dict[str, SpecDataT] = dict()
 
         # Dict of {linking text => link-defaults data}
-        self.defaultSpecs: t.DefaultDict[str, t.List[t.Dict]] = defaultdict(list)
+        self.defaultSpecs: t.LinkDefaultsT = defaultdict(list)
 
         # Set of spec vshortnames to remove from consideration when there are other possible anchors
-        self.ignoredSpecs: t.Set[str] = set()
+        self.ignoredSpecs: set[str] = set()
 
         # Set of (obsolete spec vshortname, replacing spec vshortname), when both obsolete and replacing specs are possible anchors
-        self.replacedSpecs: t.Set[t.Tuple[str, str]] = set()
+        self.replacedSpecs: set[tuple[str, str]] = set()
 
         # Dict of {biblio term => biblio data}
         # Sparsely populated, with more loaded on demand
-        self.biblios: t.DefaultDict[str, t.List[biblio.BiblioEntry]] = defaultdict(list)
-        self.loadedBiblioGroups: t.Set[str] = set()
+        self.biblios: defaultdict[str, list[biblio.BiblioEntry]] = defaultdict(list)
+        self.loadedBiblioGroups: set[str] = set()
 
         # Most of the biblio keys, for biblio near-miss correction
         # (Excludes dated versions, and all huge "foo\d+" groups that can't usefully correct typos.)
-        self.biblioKeys: t.Set[str] = set()
+        self.biblioKeys: set[str] = set()
 
         # Dict of {suffixless key => [keys with numeric suffixes]}
         # (So you can tell when it's appropriate to default a numeric-suffix ref to a suffixless one.)
-        self.biblioNumericSuffixes: t.Dict[str, t.List[str]] = dict()
+        self.biblioNumericSuffixes: dict[str, list[str]] = dict()
 
         # Dict of {base key name => preferred display name}
-        self.preferredBiblioNames: t.Dict[str, str] = dict()
+        self.preferredBiblioNames: dict[str, str] = dict()
 
         # Dict of {spec vshortname => headings}
-        # Each heading is either {#foo => heading-dict}, {/foo#bar => heading-dict} or {#foo => [page-heading-keys]}
-        # In the latter, it's a list of the heading keys (of the form /foo#bar) that collide for that id.
-        self.headings: t.Dict[str, t.Union[t.Dict[str, str], t.List[str]]] = dict()
+        self.headings: dict[str, headingdata.SpecHeadings] = dict()
 
+        self.defaultStatus: str
         if defaultStatus is None:
             self.defaultStatus = constants.refStatus.current
         else:
             self.defaultStatus = constants.refStatus[defaultStatus]
 
-        self.localRefs = source.RefSource("local", fileRequester=fileRequester)
-        self.anchorBlockRefs = source.RefSource("anchor-block", fileRequester=fileRequester)
-        self.foreignRefs = source.RefSource(
+        self.localRefs: source.RefSource = source.RefSource("local", fileRequester=fileRequester)
+        self.anchorBlockRefs: source.RefSource = source.RefSource("anchor-block", fileRequester=fileRequester)
+        self.foreignRefs: source.RefSource = source.RefSource(
             "foreign",
             specs=self.specs,
             ignored=self.ignoredSpecs,
             replaced=self.replacedSpecs,
             fileRequester=fileRequester,
         )
-        self.shortname = None
-        self.specLevel = None
-        self.spec = None
+        self.shortname: str | None = None
+        self.specLevel: str | None = None
+        self.spec: str | None = None
 
-    def initializeRefs(self, doc: t.SpecT = None) -> None:
+    def initializeRefs(self, datablocks: t.ModuleType, doc: t.SpecT | None = None) -> None:
         """
         Load up the xref data
         This is oddly split up into sub-functions to make it easier to track performance.
         """
 
-        def initSpecs():
+        def initSpecs() -> None:
             self.specs.update(json.loads(self.dataFile.fetch("specs.json", str=True)))
 
         initSpecs()
 
-        def initMethods():
-            self.foreignRefs.methods.update(json.loads(self.dataFile.fetch("methods.json", str=True)))
+        def initMethods() -> None:
+            for arglessSig, argfulls in json.loads(self.dataFile.fetch("methods.json", str=True)).items():
+                variants = source.MethodVariants(arglessSig, {})
+                self.foreignRefs.methods[arglessSig] = variants
+                for argfullSig, data in argfulls.items():
+                    variants.variants[argfullSig] = source.MethodVariant(
+                        argfullSig, data["args"], data["for"], data["shortname"]
+                    )
 
         initMethods()
 
-        def initFors():
+        def initFors() -> None:
             self.foreignRefs.fors.update(json.loads(self.dataFile.fetch("fors.json", str=True)))
 
         initFors()
@@ -169,7 +181,7 @@ class ReferenceManager:
                 except OSError:
                     m.warn("link-defaults.infotree not found despite being listed in the External Infotrees metadata.")
 
-    def fetchHeadings(self, spec):
+    def fetchHeadings(self, spec: str) -> headingdata.SpecHeadings:
         if spec in self.headings:
             return self.headings[spec]
         with self.dataFile.fetch("headings", f"headings-{spec}.json", okayToFail=True) as fh:
@@ -177,59 +189,26 @@ class ReferenceManager:
                 data = json.load(fh)
             except ValueError:
                 # JSON couldn't be decoded, *should* only be because of empty file
-                return
-            self.headings[spec] = data
-            return data
+                data = {}
+            specHeadings = headingdata.SpecHeadings(spec, data)
+            self.headings[spec] = specHeadings
+            return specHeadings
 
     def fetchHeading(
-        self, spec: str, id: str, status: t.Optional[str] = None, el: t.Optional[t.ElementT] = None
-    ) -> t.Optional[t.Dict[str, str]]:
-        headingData = self.fetchHeadings(spec)
-        # Each heading is either {#id => heading-dict}, {/page#id => heading-dict} or {#id => ["/page#id"]}
-        # In the latter, it's a list of the heading keys (of the form /foo#bar) that collide for that id.
-        currentHeading: t.Optional[t.Dict[str, str]]
-        snapshotHeading: t.Optional[t.Dict[str, str]]
-        if id in headingData:
-            if isinstance(headingData[id], dict):
-                currentHeading = headingData[id].get("current")
-                snapshotHeading = headingData[id].get("snapshot")
-            if isinstance(headingData[id], list):
-                clashingIds = headingData[id]
-                # Multipage spec, list is "/page#id" keys
-                # that (potentially) collide for that heading ID
-                if len(clashingIds) == 1:
-                    # only one heading of this name, no worries
-                    x = headingData[clashingIds[0]]
-                    currentHeading = x.get("current")
-                    snapshotHeading = x.get("snapshot")
-                else:
-                    # multiple headings of this id, user needs to disambiguate
-                    m.die(
-                        f"Multiple headings with id '{id}' for spec '{spec}'. Please specify:\n"
-                        + "\n".join(f"  [[{spec + x}]]" for x in clashingIds),
-                        el=el,
-                    )
-                    return None
-        else:
-            m.die(
-                f"Couldn't find section '{id}' in spec '{spec}':\n{h.outerHTML(el)}",
-                el=el,
-            )
-            return None
+        self, spec: str, id: str, status: str | None = None, el: t.ElementT | None = None
+    ) -> dict[str, str] | None:
+        specHeadings = self.fetchHeadings(spec)
         if status is None:
             status = self.defaultStatus
-        if status == "current":
-            return currentHeading or snapshotHeading
-        else:
-            return snapshotHeading or currentHeading
+        return specHeadings.get(id, status, el)
 
-    def initializeBiblio(self):
+    def initializeBiblio(self) -> None:
         self.biblioKeys.update(json.loads(self.dataFile.fetch("biblio-keys.json", str=True)))
         self.biblioNumericSuffixes.update(json.loads(self.dataFile.fetch("biblio-numeric-suffixes.json", str=True)))
 
         # Get local bibliography data
         try:
-            storage = defaultdict(list)
+            storage: t.BiblioStorageT = defaultdict(list)
             with open("biblio.json", encoding="utf-8") as fh:
                 biblio.processSpecrefBiblioFile(fh.read(), storage, order=2)
         except OSError:
@@ -254,7 +233,7 @@ class ReferenceManager:
             )
         )
 
-    def setSpecData(self, md):
+    def setSpecData(self, md: t.MetadataManager) -> None:
         if md.defaultRefStatus:
             self.defaultStatus = md.defaultRefStatus
         elif md.status in config.snapshotStatuses:
@@ -273,19 +252,21 @@ class ReferenceManager:
         # with the possibility of overriding the "shortname-level" pattern.
         self.removeSameSpecRefs()
 
-    def removeSameSpecRefs(self):
+    def removeSameSpecRefs(self) -> None:
         # Kill all the non-local anchors with the same shortname as the current spec,
         # so you don't end up accidentally linking to something that's been removed from the local copy.
         # TODO: This is dumb.
         for _, refs in self.foreignRefs.refs.items():
             for ref in refs:
-                if ref["status"] != "local" and ref["shortname"].rstrip() == self.shortname:
-                    ref["export"] = False
+                if ref.status != "local" and ref.shortname.rstrip() == self.shortname:
+                    ref._ref["export"] = False  # pylint: disable=protected-access
 
-    def addLocalDfns(self, dfns):
+    def addLocalDfns(self, doc: t.SpecT, dfns: t.Iterable[t.ElementT]) -> None:
         for el in dfns:
-            if h.hasClass(el, "no-ref"):
+            if h.hasClass(doc, el, "no-ref"):
                 continue
+            elId = el.get("id")
+            assert elId is not None
             try:
                 linkTexts = config.linkTextsFromElement(el)
             except config.DuplicatedLinkText as e:
@@ -303,9 +284,9 @@ class ReferenceManager:
                     continue
                 if linkType in config.lowercaseTypes:
                     linkText = linkText.lower()
-                dfnFor = h.treeAttr(el, "data-dfn-for")
-                if dfnFor is None:
-                    dfnFor = set()
+                dfnForAttr = h.treeAttr(el, "data-dfn-for")
+                if dfnForAttr is None:
+                    dfnFor: set[str] = set()
                     existingRefs = self.localRefs.queryRefs(linkType=linkType, text=linkText, linkFor="/", exact=True)[
                         0
                     ]
@@ -313,7 +294,7 @@ class ReferenceManager:
                         m.die(f"Multiple local '{linkType}' <dfn>s have the same linking text '{linkText}'.", el=el)
                         continue
                 else:
-                    dfnFor = set(config.splitForValues(dfnFor))
+                    dfnFor = set(config.splitForValues(dfnForAttr))
                     encounteredError = False
                     for singleFor in dfnFor:
                         existingRefs = self.localRefs.queryRefs(
@@ -338,26 +319,30 @@ class ReferenceManager:
                     if match:
                         dfnFor.add(match.group(1).strip())
                 # convert back into a list now, for easier JSONing
-                dfnFor = sorted(dfnFor)
-                ref = {
-                    "type": linkType,
-                    "status": "local",
-                    "spec": self.spec,
-                    "shortname": self.shortname,
-                    "level": self.specLevel,
-                    "url": "#" + el.get("id"),
-                    "export": True,
-                    "for": dfnFor,
-                    "el": el,
-                }
+                dfnForList = list(sorted(dfnFor))
+                ref = wrapper.RefWrapper(
+                    linkText,
+                    {
+                        "type": linkType,
+                        "status": "local",
+                        "spec": self.spec,
+                        "shortname": self.shortname,
+                        "level": self.specLevel,
+                        "url": "#" + elId,
+                        "export": True,
+                        "normative": True,
+                        "for_": dfnForList,
+                        "el": el,
+                    },
+                )
                 self.localRefs.refs[linkText].append(ref)
                 for for_ in dfnFor:
                     self.localRefs.fors[for_].append(linkText)
                 methodishStart = re.match(r"([^(]+\()[^)]", linkText)
                 if methodishStart:
-                    self.localRefs.addMethodVariants(linkText, dfnFor, ref["shortname"])
+                    self.localRefs.addMethodVariants(linkText, dfnForList, ref.shortname)
 
-    def filterObsoletes(self, refs):
+    def filterObsoletes(self, refs: list[RefWrapper]) -> list[RefWrapper]:
         return utils.filterObsoletes(
             refs,
             replacedSpecs=self.replacedSpecs,
@@ -366,7 +351,7 @@ class ReferenceManager:
             localSpec=self.spec,
         )
 
-    def queryAllRefs(self, **kwargs):
+    def queryAllRefs(self, **kwargs: t.Any) -> list[RefWrapper]:
         r1, _ = self.localRefs.queryRefs(**kwargs)
         r2, _ = self.anchorBlockRefs.queryRefs(**kwargs)
         r3, _ = self.foreignRefs.queryRefs(**kwargs)
@@ -377,17 +362,17 @@ class ReferenceManager:
 
     def getRef(
         self,
-        linkType,
-        text,
-        spec=None,
-        status=None,
-        statusHint=None,
-        linkFor=None,
-        explicitFor=False,
-        linkForHint=None,
-        error=True,
-        el=None,
-    ):
+        linkType: str,
+        text: str,
+        spec: str | None = None,
+        status: str | None = None,
+        statusHint: str | None = None,
+        linkFor: str | list[str] | None = None,
+        explicitFor: bool = False,
+        linkForHint: str | None = None,
+        error: bool = True,
+        el: t.ElementT | None = None,
+    ) -> RefWrapper | None:
         # If error is False, this function just shuts up and returns a reference or None
         # Otherwise, it pops out debug messages for the user.
 
@@ -454,7 +439,7 @@ class ReferenceManager:
 
         if status == "local":
             # Already checked local refs, can early-exit now.
-            return
+            return None
 
         # Take defaults into account
         if not spec or not status or not linkFor:
@@ -545,15 +530,15 @@ class ReferenceManager:
                 # Find all method signatures that contain the arg in question
                 # and, if interface is specified, are for that interface.
                 # Dedup/collect by url, so I'll get all the signatures for a given dfn.
-                possibleMethods = defaultdict(list)
-                for argfullName, metadata in methodSignatures.items():
+                possibleMethodsDict: defaultdict[str, list[str]] = defaultdict(list)
+                for variant in methodSignatures.variants.values():
                     if (
-                        text in metadata["args"]
-                        and (interfaceName in metadata["for"] or interfaceName is None)
-                        and metadata["shortname"] != self.shortname
+                        text in variant.args
+                        and (interfaceName in variant.for_ or interfaceName is None)
+                        and variant.shortname != self.shortname
                     ):
-                        possibleMethods[metadata["shortname"]].append(argfullName)
-                possibleMethods = list(possibleMethods.values())
+                        possibleMethodsDict[variant.shortname or ""].append(variant.signature)
+                possibleMethods = list(possibleMethodsDict.values())
                 if not possibleMethods:
                     # No method signatures with this argument/interface.
                     # Jump out and fail in a normal way.
@@ -619,7 +604,7 @@ class ReferenceManager:
                     f"Too many possible method targets to disambiguate '{linkFor}/{text}'. Please specify the names of the required args, like 'foo(bar, baz)', in the 'for' attribute.",
                     el=el,
                 )
-                return
+                return None
             # Otherwise
 
         if failure in ("text", "type"):
@@ -691,13 +676,13 @@ class ReferenceManager:
     def getBiblioRef(
         self,
         text: str,
-        status: t.Optional[str] = None,
+        status: str | None = None,
         generateFakeRef: bool = False,
         allowObsolete: bool = False,
-        el: t.Optional[t.ElementT] = None,
+        el: t.ElementT | None = None,
         quiet: bool = False,
         depth: int = 0,
-    ) -> t.Optional[biblio.BiblioEntry]:
+    ) -> biblio.BiblioEntry | None:
         if depth > 100:
             m.die(f"Data error in biblio files; infinitely recursing trying to find [{text}].")
             return None
@@ -787,7 +772,7 @@ class ReferenceManager:
         bib.preferredStatus = status
         return bib
 
-    def bibliosFromKey(self, key: str) -> t.List[biblio.BiblioEntry]:
+    def bibliosFromKey(self, key: str) -> list[biblio.BiblioEntry]:
         # Load up the biblio data necessary to fetch the given key
         # and then actually fetch it.
         # If you don't call this,
@@ -801,10 +786,10 @@ class ReferenceManager:
             self.loadedBiblioGroups.add(group)
         return self.biblios.get(key, [])
 
-    def _bestCandidateBiblio(self, candidates: t.List[biblio.BiblioEntry]) -> biblio.BiblioEntry:
+    def _bestCandidateBiblio(self, candidates: list[biblio.BiblioEntry]) -> biblio.BiblioEntry:
         return sorted(candidates, key=lambda x: x.order or 0)[0].strip()
 
-    def getLatestBiblioRef(self, key: str) -> t.Optional[biblio.BiblioEntry]:
+    def getLatestBiblioRef(self, key: str) -> biblio.BiblioEntry | None:
         # Takes a biblio reference name,
         # returns the latest dated variant of that name
         # (names in the form FOO-19700101)
@@ -827,7 +812,7 @@ class ReferenceManager:
             return None
         return self._bestCandidateBiblio(latestRefs)
 
-    def vNamesFromSpecNames(self, specName):
+    def vNamesFromSpecNames(self, specName: str) -> list[str]:
         # Takes an unversioned specName,
         # returns the versioned names that Shepherd knows about.
 
@@ -842,7 +827,7 @@ class ReferenceManager:
         return chosenVNames
 
 
-def simplifyPossibleRefs(refs, alwaysShowFor=False):
+def simplifyPossibleRefs(refs: list[RefWrapper], alwaysShowFor: bool = False) -> list[SimplifiedRef]:
     # "Simplifies" the set of possible refs according to their 'for' value;
     # returns a list of text/type/spec/for objects,
     # with the for value filled in *only if necessary for disambiguation*.
@@ -858,28 +843,44 @@ def simplifyPossibleRefs(refs, alwaysShowFor=False):
         if len(fors) >= 2 or alwaysShowFor:
             # Needs for-based disambiguation
             for for_, url in fors:
-                retRefs.append({"text": text, "type": type, "spec": spec, "for_": for_, "url": url})
+                retRefs.append(SimplifiedRef(text=text, type=type, spec=spec, for_=for_, url=url))
         else:
             retRefs.append(
-                {
-                    "text": text,
-                    "type": type,
-                    "spec": spec,
-                    "for_": None,
-                    "url": fors[0][1],
-                }
+                SimplifiedRef(
+                    text=text,
+                    type=type,
+                    spec=spec,
+                    for_=None,
+                    url=fors[0][1],
+                )
             )
     return retRefs
 
 
-def refToText(ref):
-    if ref["for_"]:
-        return "spec:{spec}; type:{type}; for:{for_}; text:{text}".format(**ref)
+@dataclasses.dataclass
+class SimplifiedRef:
+    text: str
+    type: str
+    spec: str
+    for_: str | None
+    url: str
+
+
+def refToText(ref: SimplifiedRef) -> str:
+    if ref.for_ is not None:
+        return f"spec:{ref.spec}; type:{ref.type}; for:{ref.for_}; text:{ref.text}"
     else:
-        return "spec:{spec}; type:{type}; text:{text}".format(**ref)
+        return f"spec:{ref.spec}; type:{ref.type}; text:{ref.text}"
 
 
-def reportMultiplePossibleRefs(possibleRefs, linkText, linkType, linkFor, defaultRef, el):
+def reportMultiplePossibleRefs(
+    possibleRefs: t.Sequence[SimplifiedRef],
+    linkText: str,
+    linkType: str,
+    linkFor: str | list[str] | None,
+    defaultRef: RefWrapper,
+    el: t.ElementT | None,
+) -> None:
     # Sometimes a badly-written spec has indistinguishable dfns.
     # Detect this by seeing if more than one stringify to the same thing.
     allRefs = defaultdict(list)
@@ -906,11 +907,13 @@ def reportMultiplePossibleRefs(possibleRefs, linkText, linkType, linkFor, defaul
         for refs in mergedRefs:
             error += "\n" + refToText(refs[0])
             for ref in refs:
-                error += "\n  " + ref["url"]
+                error += "\n  " + ref.url
     m.linkerror(error, el=el)
 
 
-def reportAmbiguousForlessLink(el, text, forlessRefs, localRefs):
+def reportAmbiguousForlessLink(
+    el: t.ElementT | None, text: str, forlessRefs: list[RefWrapper], localRefs: list[RefWrapper]
+) -> None:
     localRefText = "\n".join([refToText(ref) for ref in simplifyPossibleRefs(localRefs, alwaysShowFor=True)])
     forlessRefText = "\n".join([refToText(ref) for ref in simplifyPossibleRefs(forlessRefs, alwaysShowFor=True)])
     m.linkerror(
